from logging import DEBUG
from logging.handlers import TimedRotatingFileHandler
from typing import Optional

from delogger.handlers.count_rotating_file import CountRotatingFileHandler
from delogger.modes.base import ModeBase
from delogger.util.log_file import LogFile

__all__ = ["CountRotatingFileMode", "TimedRotatingFileMode"]


class FileModeBase(ModeBase):
    fmt = "%(asctime)s.%(msecs).03d %(levelname)s %(filename)s:%(lineno)d %(funcName)s %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"


class CountRotatingFileMode(FileModeBase):
    def __init__(
        self,
        filepath: str = "log/%Y%m%d_%H%M%S.log",
        backup_count: int = 5,
        level: int = DEBUG,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.level = level
        self.filepath = filepath
        self.backup_count = backup_count

        self.logfile: Optional[LogFile] = None

    def load(self, delogger) -> None:
        run_hdlr = CountRotatingFileHandler(
            filepath=self.filepath, backup_count=self.backup_count
        )

        delogger.add_handler(run_hdlr, self.level, fmt=self.fmt, datefmt=self.datefmt)

        self.logfile = LogFile(run_hdlr.filepath)


class TimedRotatingFileMode(FileModeBase):
    def __init__(
        self,
        filepath: str = "log/delogger.log",
        when: str = "midnight",
        backup_count: int = 0,
        level: int = DEBUG,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.level = level
        self.filepath = filepath
        self.when = when
        self.backup_count = backup_count

        self.logfile = LogFile(filepath)
        self.logfile.mkdir()

    def load(self, delogger) -> None:
        timed_hdlr = TimedRotatingFileHandler(
            filename=str(self.logfile.filepath),
            when=self.when,
            backupCount=self.backup_count,
        )

        delogger.add_handler(timed_hdlr, self.level, fmt=self.fmt, datefmt=self.datefmt)



from datetime import datetime as dt
from pathlib import Path

__all__ = ["LogFile"]


class LogFile:
    """Set the path of the log file.

    Args:
        filepath (str): Log file path.

    Attributes:
        filepath (Path): Log file path.
        filepath_raw (str): Log file path raw.

    """

    def __init__(self, filepath: str) -> None:
        self.filepath: Path = Path(dt.today().strftime(filepath))
        self.filepath_raw: Path = Path(filepath)

    def mkdir(self) -> None:
        dirpath = self.filepath.parent
        if dirpath.is_dir():
            return

        dirpath.mkdir(parents=True, exist_ok=True)

    def __eq__(self, other):
        """Comparison for CountRotatingFileHandler.

        Returns:
            True if dirpath is the same, False otherwise.

        """
        if not isinstance(other, LogFile):
            raise NotImplementedError

        return other.filepath.absolute() == self.filepath.absolute()

    def __contains__(self, other):
        return other == self

    def __str__(self):
        return str(self.filepath)



class OnlyFilter:
    """A filter that performs control to output only the specified level.

    Args:
        level (int): Handler level.

    Attributes:
        _level (int): Handler level.

    """

    def __init__(self, level):
        self._level = level

    def filter(self, record):
        return record.levelno == self._level



from logging import CRITICAL
from logging import DEBUG
from logging import Formatter
from logging import Logger
from logging import StreamHandler
from logging import WARNING
from logging import addLevelName
from logging import getLogger
import os
from typing import Dict
from typing import List
from typing import Optional

from colorlog import ColoredFormatter

from delogger.decorators.base import DecoratorBase
from delogger.filters.only_filter import OnlyFilter
from delogger.modes.base import ModeBase


class DeloggerBase:
    """A class that provides a decided logger.

    Args:
        name (str): Logger name
        parent (str): Log file save destination.
        *args: DeloggerSetting.
        **kwargs: DeloggerSetting.

    Attributes:
        _logger (logging.Logger): Logger.
        _is_new_logger (bool): Whether it is a first generation logger.

    """

    def __init__(
        self,
        name: Optional[str] = None,
        modes: Optional[List[ModeBase]] = None,
        decorators: Optional[List[DecoratorBase]] = None,
    ) -> None:
        addLevelName(WARNING, "WARN")
        addLevelName(CRITICAL, "CRIT")

        # base logger
        name = name or os.getenv("DELOGGER_NAME", "delogger")
        logger = getLogger(name)
        logger.setLevel(DEBUG)
        self._logger: Logger = logger

        # check already set logger
        if len(self._logger.handlers) > 0:
            # Already set logger
            self._is_new_logger = False
        else:
            # Not set logger
            self._is_new_logger = True

        if not self.is_already_setup():
            if modes:
                self.load_modes(*modes)

            if decorators:
                self.load_decorators(*decorators)

    def get_logger(self) -> Logger:
        if self.is_already_setup():
            return self._logger

        self._is_new_logger = False

        return self._logger

    def load_mode(self, mode: ModeBase) -> None:
        mode.load(delogger=self)

    def load_modes(self, *modes) -> None:
        for mode in modes:
            self.load_mode(mode)

    def load_decorator(self, decorator: DecoratorBase) -> None:
        decorator.load(delogger=self)

    def load_decorators(self, *decorators) -> None:
        for decorator in decorators:
            self.load_decorator(decorator)

    def is_already_setup(self) -> bool:
        return not self._is_new_logger

    def add_handler(
        self,
        hdlr,
        level: int,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        only_level: bool = False,
        formatter=None,
    ) -> None:
        """Helper function to add a handler.

        Args:
            hdlr: handler
            level (int): Handler level.
            fmt (str): Handler output format.
            datefmt (str): Handler output date format.
            only_level (bool): Whether to output only to specified han-
            dler level.
            formatter: Handler formatter.

        """

        hdlr.setLevel(level)

        # Set formatter.
        formatter = formatter or Formatter(fmt, datefmt)
        hdlr.setFormatter(formatter)

        if only_level:
            hdlr.addFilter(OnlyFilter(level))

        self._logger.addHandler(hdlr)

    def add_stream_handler(self, level: int, *, hdlr=None, **kwargs) -> None:
        """Helper function to add a stream handler.

        Args:
            level (int): Handler level.
            hdlr: Handler other than stream handler.
            **kwargs: Keyword argument of add_handler method

        """

        hdlr = hdlr or StreamHandler()
        self.add_handler(hdlr, level, **kwargs)

    def add_stream_color_handler(
        self,
        level: int,
        log_colors: Optional[Dict[str, str]],
        *,
        hdlr=None,
        datefmt: Optional[str] = None,
        **kwargs
    ) -> None:
        """Helper function to add a color stream handler.

        Args:
            level (int): Handler level.
            hdlr: Handler other than stream handler.
            **kwargs: Keyword argument of add_handler method

        """

        formatter = ColoredFormatter(
            kwargs.get("fmt", None), log_colors=log_colors, style="%", datefmt=datefmt
        )

        hdlr = hdlr or StreamHandler()
        self.add_handler(hdlr, level, datefmt=datefmt, formatter=formatter, **kwargs)

    @property
    def propagate(self) -> bool:
        return self._logger.propagate

    @propagate.setter
    def propagate(self, is_propagate: bool) -> None:
        self._logger.propagate = is_propagate



from abc import ABC
from abc import abstractmethod
from typing import Optional

__all__ = ["ModeBase"]


class ModeBase(ABC):
    fmt: str = "%(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"

    def __init__(
        self, fmt: Optional[str] = None, datefmt: Optional[str] = None
    ) -> None:
        self.fmt = fmt or self.fmt
        self.datefmt = datefmt or self.datefmt

    @abstractmethod
    def load(self, delogger) -> None:
        pass



from delogger.decorators.base import DecoratorBase
from delogger.util.warn import warn_import

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from memory_profiler import profile

    _can_memory_profiler = True
except ImportError:  # pragma: no cover
    _can_memory_profiler = False


class MemoryProfile(DecoratorBase):
    decorator_name = "memory_profile"

    def can_run(self) -> bool:
        if not _can_memory_profiler:
            warn_import(self.decorator_name, "memory_profiler")
            return False

        return True

    def wrapper(self, f, *args, **kwargs):
        # output memory_profiler
        with StringIO() as s:
            rtn = profile(f, stream=s, precision=2)(*args, **kwargs)
            msg = "memory_profiler result\n{}".format(s.getvalue())
        self.logger.debug(msg)

        return rtn



from abc import ABC
from abc import abstractmethod
from logging import Logger
from typing import Callable
from typing import Optional


class DecoratorBase(ABC):
    def __init__(self) -> None:
        self.logger: Optional[Logger] = None

    @property
    @abstractmethod
    def decorator_name(self) -> str:
        """decorator name"""

    @abstractmethod
    def wrapper(self, f, *args, **kwargs):
        """wrapper method"""

    def decorator(self, f) -> Callable:
        def wrapper(*args, **kwargs):
            if not self.can_run():
                return f(*args, **kwargs)
            return self.wrapper(f, *args, **kwargs)

        return wrapper

    def load(self, delogger) -> None:
        logger = delogger._logger
        _decorator = getattr(logger, self.decorator_name, None)
        if _decorator:
            raise AttributeError(f"'{self.decorator_name}' already defined.")

        self.logger = logger
        setattr(logger, self.decorator_name, self.decorator)

    def can_run(self) -> bool:
        return True



from logging import DEBUG
from logging import INFO
from typing import Dict
from typing import Optional

from delogger.modes.base import ModeBase

__all__ = [
    "StreamColorDebugMode",
    "StreamDebugMode",
    "StreamInfoMode",
]


class StreamModeBase(ModeBase):
    pass


class StreamDebugMode(StreamModeBase):
    fmt = "%(levelname)-5s | %(message)s"

    def load(self, delogger) -> None:
        delogger.add_stream_handler(DEBUG, fmt=self.fmt, datefmt=self.datefmt)


class StreamInfoMode(StreamModeBase):
    fmt = "%(message)s"

    def load(self, delogger) -> None:
        delogger.add_stream_handler(INFO, fmt=self.fmt, datefmt=self.datefmt)


class StreamColorModeBase(StreamModeBase):
    log_colors = {
        "DEBUG": "cyan",
        "INFO": "green",
        "WARN": "yellow",
        "ERROR": "red",
        "CRIT": "red,bg_white",
    }
    """Definition of color stream level setting."""

    def __init__(self, log_colors: Optional[Dict[str, str]] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        self.log_colors = log_colors or self.log_colors


class StreamColorDebugMode(StreamColorModeBase):
    fmt = "%(log_color)s%(levelname)-5s%(reset)s | %(message)s"

    def load(self, delogger) -> None:
        delogger.add_stream_color_handler(
            DEBUG, log_colors=self.log_colors, fmt=self.fmt, datefmt=self.datefmt
        )



from typing import Generator
from typing import List
from typing import Tuple

from delogger.decorators.base import DecoratorBase
from delogger.util.warn import warn_import

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from line_profiler import LineProfiler

    _can_line_profiler = True
except ImportError:  # pragma: no cover
    _can_line_profiler = False


try:
    from memory_profiler import profile

    _can_memory_profiler = True
except ImportError:  # pragma: no cover
    _can_memory_profiler = False


class LineEmpty(Exception):
    pass


class LineMemoryProfile(DecoratorBase):
    decorator_name = "line_memory_profile"

    L_M_HEADER_INDEX: int = 6
    L_M_SEPARATOR_INDEX: int = 7
    L_M_TEMPLATE: str = "{0:>6} {1:>9} {2:>12} {3:>8} {4:>8} " "{5:>12} {6:>12}   {7:<}"
    L_M_HEADER: List[str] = [
        "Line #",
        "Hits",
        "Time",
        "Per Hit",
        "% Time",
        "Mem usage",
        "Increment",
        "Line Contents",
    ]

    def can_run(self) -> bool:
        if not _can_line_profiler:
            warn_import(self.decorator_name, "line_profiler")
            return False

        if not _can_memory_profiler:
            warn_import(self.decorator_name, "memory_profiler")
            return False

        return True

    def wrapper(self, f, *args, **kwargs):
        # memory_profiler
        with StringIO() as s:
            rtn = profile(f, stream=s, precision=2)(*args, **kwargs)
            memory_value = self._memory_profiler_parse(s.getvalue())

        # line_profiler
        prof = LineProfiler()
        prof.add_function(f)

        rtn = prof.runcall(f, *args, **kwargs)
        with StringIO() as s:
            prof.print_stats(stream=s)
            mix, line_tmp = self._line_profiler_parse(s.getvalue())

        # memory line mix output
        template = self.L_M_TEMPLATE
        for l, m in zip(line_tmp, memory_value):
            l_m_mix = l[:5] + m
            mix.append(template.format(*l_m_mix))
        mix[self.L_M_HEADER_INDEX] = template.format(*self.L_M_HEADER)
        mix[self.L_M_SEPARATOR_INDEX] += "=" * 27
        self.logger.debug("line, memory profiler result\n" + "\n".join(mix))

        return rtn

    @staticmethod
    def _memory_profiler_parse(result) -> Generator[List[str], None, None]:
        for line in result.split("\n"):
            try:
                elem = line.split()
                if not elem:
                    continue

                lineno = elem[0].isnumeric()
                if not lineno:
                    continue

                contents = line[35:]
                if len(elem) < 2:
                    raise LineEmpty()
                float(elem[1])

                mem = elem[1:3]
                inc_mem = elem[3:5]
                yield [" ".join(mem), " ".join(inc_mem), contents]

            except ValueError:
                yield ["", "", contents]
                continue
            except LineEmpty:
                yield ["", "", ""]
                continue

    @staticmethod
    def _line_profiler_parse(result) -> Tuple[List[str], List[List[str]]]:
        mix: List[str] = []
        line_tmp: List[List[str]] = []
        for line in result.split("\n"):
            try:
                elem = line.split()
                if not elem:
                    mix.append(line)
                    continue

                is_lineno = elem[0].isnumeric()
                if not is_lineno:
                    mix.append(line)
                    continue

                if len(elem) < 2:
                    raise LineEmpty()
                float(elem[1])
                line_tmp.append(elem[:5])

            except ValueError:
                line_tmp.append([elem[0], "", "", "", ""])
                continue
            except LineEmpty:
                line_tmp.append([elem[0], "", "", "", ""])
                continue

        for _ in range(5):
            if not mix[-1]:
                del mix[-1]
            else:
                break

        return mix, line_tmp



from logging import INFO
from typing import Optional

from delogger.handlers.slack import SlackHandler
from delogger.modes.base import ModeBase

__all__ = ["SlackWebhookMode", "SlackTokenMode"]


class SlackWebhookMode(ModeBase):
    def __init__(self, webhook_url: str, level: Optional[int] = INFO, **kwargs) -> None:
        super().__init__(**kwargs)

        self.webhook_url = webhook_url
        self.level = level

    def load(self, delogger) -> None:
        slack_hdlr = SlackHandler(url=self.webhook_url, as_user=True)
        delogger.add_handler(slack_hdlr, self.level, self.fmt, self.datefmt)


class SlackTokenMode(ModeBase):
    def __init__(
        self, token: str, channel: str, level: Optional[int] = INFO, **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.token = token
        self.channel = channel
        self.level = level

    def load(self, delogger) -> None:
        slack_hdlr = SlackHandler(token=self.token, channel=self.channel, as_user=True)
        delogger.add_handler(slack_hdlr, self.level, self.fmt, self.datefmt)



from delogger.decorators.line_memory_profile import LineMemoryProfile
from delogger.decorators.line_profile import LineProfile
from delogger.decorators.line_profile import LineProfileStats
from delogger.decorators.memory_profile import MemoryProfile

__all__ = ["LineProfile", "MemoryProfile", "LineMemoryProfile", "LineProfileStats"]



from logging.handlers import QueueHandler
from logging.handlers import QueueListener


class DeloggerQueueHandler(QueueHandler):
    def __init__(self, listener: QueueListener, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.listener = listener
        self.start()

    def start(self) -> None:
        self.listener.start()

    def join(self) -> None:
        self.listener.queue.join()  # type: ignore

    def close(self, *args, **kwargs):
        if self.listener._thread:  # pragma: no cover
            self.listener.stop()  # pragma: no cover

        super().close(*args, **kwargs)  # pragma: no cover



from logging import Logger

from delogger.decorators.debug_log import DebugLog
from delogger.loggers.base import DeloggerBase
from delogger.modes.stream import StreamColorDebugMode
from delogger.presets.base import PresetsBase


class DebugStreamPresets(PresetsBase):
    def make_logger(self, delogger: DeloggerBase) -> Logger:
        delogger.load_modes(StreamColorDebugMode())
        delogger.load_decorators(DebugLog())

        return delogger.get_logger()


logger = DebugStreamPresets("debug_stream").get_logger()



from abc import ABC
from abc import abstractmethod
from logging import Logger
from os import getenv
from typing import Optional

from delogger import Delogger
from delogger import DeloggerQueue
from delogger.loggers.base import DeloggerBase
from delogger.modes.file import CountRotatingFileMode
from delogger.modes.file import TimedRotatingFileMode
from delogger.modes.slack import SlackWebhookMode


class PresetsBase(ABC):
    def __init__(self, name: Optional[str] = None, is_queue: bool = False) -> None:
        self.name = getenv("DELOGGER_NAME", name)
        self.is_queue = is_queue

        self.filepath: Optional[str] = getenv("DELOGGER_FILEPATH")
        self.slack_webhook: Optional[str] = getenv("DELOGGER_SLACK_WEBHOOK")

    @abstractmethod
    def make_logger(self, delogger: DeloggerBase) -> Logger:
        pass

    def get_logger(self) -> Logger:
        _delogger = DeloggerQueue if self.is_queue else Delogger
        return self.make_logger(_delogger(self.name))

    def count_rorating_filemode(self) -> CountRotatingFileMode:
        return (
            CountRotatingFileMode(self.filepath)
            if self.filepath
            else CountRotatingFileMode()
        )

    def timed_rotating_filemode(self) -> TimedRotatingFileMode:
        return (
            TimedRotatingFileMode(self.filepath)
            if self.filepath
            else TimedRotatingFileMode()
        )

    def slack_webhook_mode(self) -> Optional[SlackWebhookMode]:
        slack_webhook = self.slack_webhook
        if slack_webhook:
            return SlackWebhookMode(slack_webhook)

        return None



from logging import FileHandler
import os
from pathlib import Path
import re
from typing import Dict
from typing import List

from delogger.util.log_file import LogFile

__all__ = ["CountRotatingFileHandler"]


class CountRotatingFileHandler(FileHandler):
    """This handler to keep the saved log count.

    Args:
        filepath (str): log filepath.
        backup_count (int): Leave logs up to the designated generation.

    Attributes:
        filepath (str): File path determined only once at runtime.

    """

    _files: List[LogFile] = []
    """This list saving LogFile of output log file."""

    _LOG_FMT_RE: Dict[str, List[str]] = {
        r"\d{4}": ["%Y"],
        r"\d{2}": ["%m", "%d", "%H", "%M", "%S"],
        r"\d{6}": ["%f"],
    }

    def __init__(self, filepath: str, backup_count: int = 5) -> None:
        dirpath = str(Path(filepath).parent)
        fmt = Path(filepath).name

        self.logfile = self._load_file_path(dirpath, fmt, backup_count)
        self.filepath = str(self.logfile.filepath)

        super().__init__(self.filepath)

    def _open(self):
        """It is executed at log output.

        If there is no directory, it will be created automatically.

        """

        # If there is no save destination directory, the directory is created.
        self.logfile.mkdir()

        return super()._open()

    def _load_file_path(self, dirpath: str, fmt: str, backup_count: int) -> LogFile:
        """Get the file path of the log output destination.

        For each directory, determine the log file path only once at runtime.

        Args:
            dirpath (str): Directory path.
            fmt (str): Filename like date_string.
            backup_count (int): Leave logs up to the designated generation.

        """

        logfile = LogFile(str(Path(dirpath) / fmt))

        # If already same logfile, return the filepath.
        for fpath in CountRotatingFileHandler._files:
            if fpath == logfile:
                return fpath
        CountRotatingFileHandler._files.append(logfile)

        if backup_count <= 0:
            return logfile

        file_list = self._get_match_files(logfile.filepath.parent, fmt)

        # Delete the old file and set a new filepath
        if len(file_list) >= backup_count:
            os.remove(file_list[0])
        return logfile

    def _get_match_files(self, dirpath, fmt) -> List[Path]:
        _fmt = fmt
        for patter, date_strs in self._LOG_FMT_RE.items():
            for date_str in date_strs:
                _fmt = _fmt.replace(date_str, patter)

        # TODO: check file stat?? datefmt only ok??
        pattern = re.compile(_fmt)
        files = [x for x in sorted(Path(dirpath).glob("*")) if pattern.search(str(x))]

        return files



from logging import Logger

from delogger.decorators.debug_log import DebugLog
from delogger.decorators.profiles import LineMemoryProfile
from delogger.decorators.profiles import LineProfile
from delogger.decorators.profiles import LineProfileStats
from delogger.decorators.profiles import MemoryProfile
from delogger.loggers.base import DeloggerBase
from delogger.modes.stream import StreamColorDebugMode
from delogger.presets.base import PresetsBase


class ProfilerPresets(PresetsBase):
    def make_logger(self, delogger: DeloggerBase) -> Logger:
        delogger.load_modes(StreamColorDebugMode(), self.count_rorating_filemode())
        delogger.load_decorators(
            DebugLog(),
            LineProfile(),
            LineProfileStats(),
            MemoryProfile(),
            LineMemoryProfile(),
        )

        return delogger.get_logger()


logger = ProfilerPresets("profiler_logger").get_logger()



from .decorators.base import DecoratorBase
from .loggers.delogger import Delogger
from .loggers.delogger_queue import DeloggerQueue
from .modes.base import ModeBase

__all__ = (
    "Delogger",
    "DeloggerQueue",
    "ModeBase",
    "DecoratorBase",
)



from logging import Logger
from logging import NOTSET
from logging.handlers import QueueListener
from queue import Queue
from typing import Optional

from delogger.handlers.delogger_queue import DeloggerQueueHandler
from delogger.loggers.base import DeloggerBase


class DeloggerQueue(DeloggerBase):
    """Non-blocking Delogger using DeloggerQueueHandler.

    Args:
        default (bool): Whether to use the default handler.
    """

    _queue_hdlr: Optional[DeloggerQueueHandler] = None
    """A common QueueListener for all loggers."""

    def get_logger(self) -> Logger:
        if not self.is_already_setup():
            self.queue_logger()
        else:
            self._queue_hdlr: Optional[DeloggerQueueHandler] = self._find_queue_hdlr(
                self._logger.handlers
            )

        return super().get_logger()

    def queue_logger(self) -> None:
        """Set up DeloggerQueueHandler.

        Set QueueListener only for the first time.

        """
        # init que and listener
        handlers = self._logger.handlers[:]
        for hdlr in handlers:
            self._logger.removeHandler(hdlr)

        que: Queue = Queue(-1)
        listener = QueueListener(que, *handlers, respect_handler_level=True)
        queue_handler = DeloggerQueueHandler(listener, que)
        self.add_handler(queue_handler, NOTSET)

        self._queue_hdlr = queue_handler

    def join(self) -> bool:
        if not self._queue_hdlr:
            return False

        self._queue_hdlr.join()
        return True

    def _find_queue_hdlr(self, handlers) -> Optional[DeloggerQueueHandler]:
        for handler in handlers:
            if isinstance(handler, DeloggerQueueHandler):
                return handler

        return None



import warnings


def warn_import(decorator, package):
    _warn(
        f"Not work correctory {decorator} decorator. Please install {package} package."
    )


def _warn(msg):
    warnings.warn(msg, ImportWarning)



import json
from logging import CRITICAL
from logging import DEBUG
from logging import ERROR
from logging import Handler
from logging import INFO
from logging import NOTSET
from logging import WARNING
from typing import Any
from typing import Dict
from typing import Optional
import urllib.request


class SlackHandler(Handler):
    """Handler to send to Slack.

    Args:
        url (str): slack webhook url.
        channel (str): Slack channel to be transmitted.
        as_user (bool): Whether to send as a user.
        token (str): slack token.

        emoji (str): Emoji on sending.
        username (str): Username on sending.
        emojis (dict): Emojis on sending.
        usernames (dict): Usernames on sending.

    Attributes:
        url (str): slack webhook url.
        channel (str): Slack channel to be transmitted.
        as_user (bool): Whether to send as a user.
        token (str): slack token.

        emoji (str): Emoji on sending.
        username (str): Username on sending.
        emojis (dict): Emojis on sending.
        usernames (dict): Usernames on sending.

    """

    TIMEOUT: int = 20
    """default timeout for request."""

    EMOJIS: Dict[int, str] = {
        NOTSET: ":loudspeaker:",
        DEBUG: ":simple_smile:",
        INFO: ":smile:",
        WARNING: ":sweat:",
        ERROR: ":sob:",
        CRITICAL: ":scream:",
    }
    """Default value of emojis."""

    USERNAMES: Dict[int, str] = {
        NOTSET: "Notset",
        DEBUG: "Debug",
        INFO: "Info",
        WARNING: "Warning",
        ERROR: "Erorr",
        CRITICAL: "Critical",
    }
    """Default value of usernames."""

    POST_MESSAGE_URL: str = "https://slack.com/api/chat.postMessage"
    """Execution when using token API."""

    def __init__(
        self,
        url: Optional[str] = None,
        channel: Optional[str] = None,
        as_user: bool = False,
        token: Optional[str] = None,
        *,
        emoji: Optional[str] = None,
        username: Optional[str] = None,
        emojis: Optional[Dict[int, str]] = None,
        usernames: Optional[Dict[int, str]] = None,
    ):
        self.is_emit = True

        self.url = url
        token = token
        if token:
            self.url = self.POST_MESSAGE_URL

        if not self.url:
            self.is_emit = False
            raise ValueError("Not set url")

        self.token = token
        self.channel = channel
        self.usernames = usernames if token else self.USERNAMES
        self.emojis = emojis if token else self.EMOJIS
        self.as_user = as_user
        self.emoji = emoji
        self.username = username

        self.headers = {"Content-Type": "application/json; charset=utf-8"}
        if token:
            self.headers["Authorization"] = f"Bearer {token}"

        super().__init__()

    def make_json(self, record) -> Dict[str, Any]:
        json_data: Dict[str, Any] = {}

        json_data["text"] = self.format(record)
        if self.as_user:
            json_data["as_user"] = self.as_user

        else:
            if self.emoji:
                json_data["icon_emoji"] = self.emoji
            elif self.emojis:
                json_data["icon_emoji"] = self.emojis[record.levelno]

            if self.username:
                json_data["username"] = self.username
            elif self.usernames:
                json_data["username"] = self.usernames[record.levelno]

        if self.channel:
            json_data["channel"] = self.channel

        return json_data

    def make_payload(self, record) -> bytes:
        """Get slack's payload."""

        json_data = self.make_json(record)

        return json.dumps(json_data).encode("utf-8")

    def emit(self, record):
        """Send a message to Slack."""

        try:
            if not self.is_emit:
                return

            payload = self.make_payload(record)
            request = urllib.request.Request(
                url=self.url, method="POST", data=payload, headers=self.headers
            )
            urllib.request.urlopen(request, timeout=self.TIMEOUT)

        except Exception:
            self.handleError(record)

    def __eq__(self, other):
        """Comparison for SlackHandler.

        Returns:
            True if token, url, level and channel is the same, False otherwise.

        """

        if not isinstance(other, SlackHandler):
            if isinstance(other, Handler):
                return False

            raise NotImplementedError

        eq = False
        if (
            other.token == self.token
            and other.url == self.url
            and other.level == self.level
            and other.channel == self.channel
        ):
            eq = True

        return eq

    def __hash__(self):
        return hash("{}{}{}{}".format(self.token, self.url, self.level, self.channel))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['delogger',
 'delogger.decorators',
 'delogger.filters',
 'delogger.handlers',
 'delogger.loggers',
 'delogger.modes',
 'delogger.presets',
 'delogger.util']

package_data = \
{'': ['*']}

install_requires = \
['colorlog>=4.2.1,<5.0.0']

setup_kwargs = {
    'name': 'delogger',
    'version': '0.3.0',
    'description': 'delogger is a convenient logging package',
    'long_description': '# delogger\n\n[![PyPI](https://badge.fury.io/py/delogger.svg)](https://badge.fury.io/py/delogger)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://github.com/deresmos/delogger/blob/master/LICENSE)\n![test](https://github.com/deresmos/delogger/workflows/Python%20Test/badge.svg)\n![CodeQL](https://github.com/deresmos/delogger/workflows/CodeQL/badge.svg)\n\n## About\n\nDelogger is a Python package that makes easy use of decided logging.\n\n### Delogger\n\n- It behaves like normal logging.\n\n### DeloggerQueue\n\n- Non-blocking logging using QueueHandler.\n\n## Installation\n\nTo install Delogger, use pip.\n\n```bash\npip install delogger\n\n```\n\n## Sample\n\n### Debug stream and output log\n\n```python\nfrom delogger.presets.debug import logger\n\nif __name__ == "__main__":\n    logger.debug("debug msg")\n    logger.info("info msg")\n    logger.warning("warn msg")\n    logger.error("error msg")\n    logger.critical("critical msg")\n```\n\nOutput\n![debug_sample_output]()\n\nFile output (./log/20201010_164622.log)\n```\n2020-10-10 16:46:22.050 DEBUG debug_preset.py:4 <module> debug msg\n2020-10-10 16:46:22.050 INFO debug_preset.py:5 <module> info msg\n2020-10-10 16:46:22.051 WARN debug_preset.py:6 <module> warn msg\n2020-10-10 16:46:22.051 ERROR debug_preset.py:7 <module> error msg\n2020-10-10 16:46:22.051 CRIT debug_preset.py:8 <module> critical msg\n```\n\n\nand more [samples](https://github.com/deresmos/delogger/tree/main/tests)\n\n## Preset\n\n- `debug`: Output color debug log and save log file.\n- `debug_stream`: Output color debug log.\n- `output`: Save log file and notify to slack.\n- `profiler`: Same debug preset and seted profiles decorator.\n\n## Mode\n\n- `CountRotatingFileMode`: Backup count rotating.\n- `TimedRotatingFileMode`: Same logging.handlers.TimedRotatingFileHandler.\n- `SlackWebhookMode`: Log to slack. (Incomming webhook)\n- `SlackTokenMode`: Log to slack. (token key)\n- `StreamColorDebugMode`: Output color log. (debug and above)\n- `StreamDebugMode`: Output noncolor log. (debug and above)\n- `StreamInfoMode`: Output noncolor log. (info and above)\n- `PropagateMode`: Set Setropagate true.\n\n## Environment\n\n- `DELOGGER_NAME`: logger name for presets.\n- `DELOGGER_FILEPATH`: output log filepath for presets.\n- `DELOGGER_SLACK_WEBHOOK`: send slack for presets.\n\n## Decorator\n\nInject decorator into logger.\n\n### debuglog\n\n```text\nDEBUG 21:00:00 debug_log.py:32 START test args=(\'test\',) kwargs={}\nDEBUG 21:00:01 debug_log.py:39 END test return=value\n```\n\n### line_profile\n\nRequired [line_profiler](https://github.com/pyutils/line_profiler) package.\n\n```text\nDEBUG 21:38:22 line_profile.py:28 line_profiler result\nTimer unit: 1e-06 s\n\nTotal time: 6.4e-05 s\nFile: test.py\nFunction: test at line 6\n\nLine #      Hits         Time  Per Hit   % Time  Line Contents\n==============================================================\n     6                                           @logger.line_profile\n     7                                           def test(arg1):\n     8       101         43.0      0.4     67.2      for i in range(100):\n     9       100         21.0      0.2     32.8          pass\n    10         1          0.0      0.0      0.0      return i\n```\n\n### memory_profile\n\nRequired [memory_profiler](https://github.com/pythonprofilers/memory_profiler) package.\n\n```text\nDEBUG 21:40:31 memory_profile.py:25 memory_profiler result\nFilename: test.py\n\nLine #    Mem usage    Increment   Line Contents\n================================================\n     6    37.96 MiB    37.96 MiB   @logger.memory_profile\n     7                             def test(arg1):\n     8    45.43 MiB     7.47 MiB       a = [0] * 1000 * 1000\n     9    45.43 MiB     0.00 MiB       for i in range(100):\n    10    45.43 MiB     0.00 MiB           pass\n    11    45.43 MiB     0.00 MiB       return i\n```\n\n### line_memory_profile\n\nRequired [line_profiler](https://github.com/pyutils/line_profiler) and [memory_profiler](https://github.com/pythonprofilers/memory_profiler) package.\n\n```text\nDEBUG 21:41:08 line_memory_profile.py:70 line, memory profiler result\nTimer unit: 1e-06 s\n\nTotal time: 0.004421 s\nFile: test.py\nFunction: test at line 6\n\nLine #      Hits         Time  Per Hit   % Time    Mem usage    Increment   Line Contents\n=========================================================================================\n     6                                             37.96 MiB    37.96 MiB   @logger.line_memory_profile\n     7                                                                      def test(arg1):\n     8         1       4355.0   4355.0     98.5    45.43 MiB     7.47 MiB       a = [0] * 1000 * 1000\n     9       101         33.0      0.3      0.7    45.43 MiB     0.00 MiB       for i in range(100):\n    10       100         32.0      0.3      0.7    45.43 MiB     0.00 MiB           pass\n    11         1          1.0      1.0      0.0    45.43 MiB     0.00 MiB       return i\n```\n\n### add_line_profile\n\n- It can adjust the timing of profile output\n\nRequired [line_profiler](https://github.com/pyutils/line_profiler) package.\n\n```text\nDEBUG 21:45:55 line_profile.py:67 line_profiler_stats result\nTimer unit: 1e-06 s\n\nTotal time: 0.009081 s\nFile: test.py\nFunction: test at line 6\n\nLine #      Hits         Time  Per Hit   % Time  Line Contents\n==============================================================\n     6                                           @logger.add_line_profile\n     7                                           def test(arg1):\n     8         2       8957.0   4478.5     98.6      a = [0] * 1000 * 1000\n     9       202         71.0      0.4      0.8      for i in range(100):\n    10       200         52.0      0.3      0.6          pass\n    11         2          1.0      0.5      0.0      return i\n```\n',
    'author': 'deresmos',
    'author_email': 'deresmos@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/deresmos/delogger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from delogger.decorators.base import DecoratorBase
from delogger.util.warn import warn_import

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from line_profiler import LineProfiler

    _can_line_profiler = True
except ImportError:  # pragma: no cover
    _can_line_profiler = False


class LineProfile(DecoratorBase):
    decorator_name = "line_profile"

    def can_run(self) -> bool:
        if not _can_line_profiler:
            warn_import(self.decorator_name, "line_profiler")
            return False

        return True

    def wrapper(self, f, *args, **kwargs):
        prof = LineProfiler()
        prof.add_function(f)

        rtn = prof.runcall(f, *args, **kwargs)
        with StringIO() as f:
            prof.print_stats(stream=f)
            msg = "line_profiler result\n{}".format(f.getvalue())
            self.logger.debug(msg)

        return rtn


class LineProfileStats(DecoratorBase):
    decorator_name = "add_line_profile"

    def __init__(self) -> None:
        super().__init__()

        self.prof: Optional[LineProfiler] = None
        if _can_line_profiler:
            self.prof = LineProfiler()

    def can_run(self) -> bool:
        if not _can_line_profiler:
            warn_import(self.decorator_name, "line_profiler")
            return False

        return True

    def wrapper(self, f, *args, **kwargs):
        self.prof.add_function(f)

        rtn = self.prof.runcall(f, *args, **kwargs)

        return rtn

    def print_stats(self) -> None:
        if not self.prof:
            return None

        with StringIO() as f:
            self.prof.print_stats(stream=f)
            msg = "line_profiler_stats result\n{}".format(f.getvalue())
            self.logger.debug(msg)

    def load(self, delogger) -> None:
        super().load(delogger)

        setattr(delogger._logger, "print_stats", self.print_stats)



from typing import Optional

from delogger.decorators.base import DecoratorBase


class DebugLog(DecoratorBase):
    decorator_name = "debuglog"

    START_MESSAGE: str = "START {func_name} args={args} kwargs={kwargs}"
    END_MESSAGE: str = "END {func_name} return={return}"

    def __init__(
        self, start_message: Optional[str] = None, end_message: Optional[str] = None
    ) -> None:
        super().__init__()

        self.start_message: str = start_message or self.START_MESSAGE
        self.end_message: str = end_message or self.END_MESSAGE

    def wrapper(self, f, *args, **kwargs):
        # Output function name and argument.
        msg = self.start_message.format(
            **{"func_name": f.__qualname__, "args": args, "kwargs": kwargs}
        )
        self.logger.debug(msg)

        # Output function name and return value.
        rtn = f(*args, **kwargs)
        msg = self.end_message.format(**{"func_name": f.__qualname__, "return": rtn})
        self.logger.debug(msg)

        return rtn



from logging import Logger

from delogger.decorators.debug_log import DebugLog
from delogger.loggers.base import DeloggerBase
from delogger.presets.base import PresetsBase


class OutputPresets(PresetsBase):
    def make_logger(self, delogger: DeloggerBase) -> Logger:
        delogger.load_modes(self.timed_rotating_filemode())
        delogger.load_decorators(DebugLog())

        slack_webhook_mode = self.slack_webhook_mode()
        if slack_webhook_mode:
            delogger.load_modes(slack_webhook_mode)

        return delogger.get_logger()


logger = OutputPresets("output_logger", is_queue=True).get_logger()
